package anywheresoftware.b4a.agraham.gpstoosgb;

import anywheresoftware.b4a.BA.Author;
import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.BA.Version;

/**
*This is the GPStoOSGB object that can make transformations between WGS84 latitude,
*longitude and altitude to OSGB36 latitude, longitude and ODN height and to National Grid
*coordinates. 
*
*The methods use the Ordnance Survey specified 7 parameter Helmert transformation.
*See http://www.ordnancesurvey.co.uk/oswebsite/gps/information/coordinatesystemsinfo/guidecontents/index.html
*While the absolute accuracy of conversion is stated by the OS to be within 5 metres the differential
*accuracy is likely to be much higher.
*
*Easting and Northing are the displacement in metres of a point from the origin of the National Grid
*For a Grid reference the 100 km square in which the point is located is allocated a two character
*reference and the location of the point in that sqare is given by 2,3,4, 5 or even 6 numerals.
*The E/W location is always stated first.
*/
@ShortName("GPStoOSGB")
@Author("Andrew Graham")
@Version(1.0f)
public class GPStoOSGB
{
	// 1.0	initial release
	private  static final double version = 1.0;

	/**
	 *The Ordnance Survey (OS) is Great Britain's national mapping agency. OS maps are
	 *based upon the OSGB36 ellipsoid and latitude and longitude shown on GB OS maps are
	 *those of the OSGB36 ellipsoid. However the OS National Grid Reference is the primary
	 *way that map locations are specified in the UK. The National Grid is printed on all
	 *OS maps as a grid as the primary location mechanism. Lat/long is included but only
	 *as small crosses at spot points for cross reference to the grid.
	 *
	 *Latitude and longitude from GPS is usually that of the WSG84 ellipsoid. This differs
	 *slightly from the OSGB36 ellipsoid.
	 *
	 *Heights shown on GB OS maps are relative to mean sea level at the Newlyn Tidal Observatory,
	 *known as the Ordnance Datum Newlyn (ODN) while altitude from GPS is usually relative to the
	 *WGS84 ellipsoid.These altitudes differ across Great Britain by between 41 to 53 metres.
	 *
	 *This library contains a GPStoOSGB object that can make transformations between WGS84 latitude,
	 *longitude and altitude and OSGB36 latitude, longitude and ODN height, and to and from National Grid
	 *coordinates.  
	 */
	public void LIBRARY_DOC()
	{
	}
	
	
	// private fields
    private static double deg2rad = Math.PI / 180;
    private static double rad2deg = 180.0 / Math.PI;
    private double deciAlt;
    private double deciLat;
    private double deciLon;
    private double easting;
    private double northing;


    // constructor
    public GPStoOSGB()
    {
    }

	/**
	 *Converts a Northing and an Easting to a National Grid Reference.
	 *Doesn't change any properties.
	 */
	public String NEtoNGR(double easting, double northing)
	{
		return NE2NGR(easting, northing);
	}

	/**
	 *Converts an OS national grid reference to OSGB36 decimal lat and long.
	 *Returns a comma separated string.
	 *The grid reference can be 2, 3, 4 or  5 digits precision, spaces are ignored
	 *Also converts it to Northing and Easting which can be retrieved
	 *from the Northing and Easting properties.
	 */
    public String NGRtoOSGB36(String ngr)
    {
        ngr = ngr.replace(" ", "");
        if (ngr.length() > 12 || ngr.length() <= 4 || ngr.length() % 2 != 0
            || !Character.isLetter(ngr.charAt(0)) || !Character.isLetter(ngr.charAt(1)))
            return "Invalid grid reference";
        int digits = (ngr.length() -2) / 2 ;
        double metres = Math.pow(10, 5 - digits);
        double north = Double.parseDouble(ngr.substring(2 + digits, 2 + digits + digits )) * metres;
        double east = Double.parseDouble(ngr.substring(2, 2 + digits)) * metres;

        double t1 = Character.toUpperCase(ngr.charAt(0)) - 65;
        if (t1 > 8)
            t1 = t1 - 1;
        double t2 = Math.floor(t1 / 5);
        north = north + 500000 * (3 - t2);
        east = east + 500000 * (t1 - 5 * t2 - 2);

        t1 = Character.toUpperCase(ngr.charAt(1)) - 65;
        if (t1 > 8)
            t1 = t1 - 1;
        t2 = Math.floor(t1 / 5);
        north = north + 100000 * (4 - t2);
        east = east + 100000 * (t1 - 5 * t2);

        easting = east;
        northing = north;


         // converts NGR easting and nothing to lat, lon.
         // input metres, output radians
        double a = 6377563.396;       // OSGB semi-major
        double b = 6356256.91;        // OSGB semi-minor
        double e0 = 400000;           // OSGB easting of false origin
        double n0 = -100000;          // OSGB northing of false origin
        double f0 = 0.9996012717;     // OSGB scale factor on central meridian
        double e2 = 0.0066705397616;  // OSGB eccentricity squared
        double lam0 = -0.034906585039886591;  // OSGB false east
        double phi0 = 0.85521133347722145;    // OSGB false north
        double af0 = a * f0;
        double bf0 = b * f0;
        double n = (af0 - bf0) / (af0 + bf0);
        double Et = east - e0;
        double phid = InitialLat(north, n0, af0, phi0, n, bf0);
        double nu = af0 / (Math.sqrt(1 - (e2 * (Math.sin(phid) * Math.sin(phid)))));
        double rho = (nu * (1 - e2)) / (1 - (e2 * (Math.sin(phid)) * (Math.sin(phid))));
        double eta2 = (nu / rho) - 1;
        double tlat2 = Math.tan(phid) * Math.tan(phid);
        double tlat4 = Math.pow(Math.tan(phid), 4);
        double tlat6 = Math.pow(Math.tan(phid), 6);
        double clatm1 = Math.pow(Math.cos(phid), -1);
        double VII = Math.tan(phid) / (2 * rho * nu);
        double VIII = (Math.tan(phid) / (24 * rho * (nu * nu * nu))) * (5 + (3 * tlat2) + eta2 - (9 * eta2 * tlat2));
        double IX = ((Math.tan(phid)) / (720 * rho * Math.pow(nu, 5))) * (61 + (90 * tlat2) + (45 * Math.pow(Math.tan(phid), 4) ));
        double phip = (phid - ((Et * Et) * VII) + (Math.pow(Et, 4) * VIII) - (Math.pow(Et, 6) * IX)); 
        double X = Math.pow(Math.cos(phid), -1) / nu;
        double XI = (clatm1 / (6 * (nu * nu * nu))) * ((nu / rho) + (2 * (tlat2)));
        double XII = (clatm1 / (120 * Math.pow(nu, 5))) * (5 + (28 * tlat2) + (24 * tlat4));
        double XIIA = clatm1 / (5040 * Math.pow(nu, 7)) * (61 + (662 * tlat2) + (1320 * tlat4) + (720 * tlat6));
        double lambdap = (lam0 + (Et * X) - ((Et * Et * Et) * XI) + (Math.pow(Et, 5) * XII) - (Math.pow(Et, 7) * XIIA));
        phip = phip * rad2deg;
        lambdap = lambdap * rad2deg;
        deciLat = phip;
        deciLon = lambdap;
        return Double.toString(phip) + ',' + Double.toString(lambdap);
    } 

    /**
     *Converts OSGB36 decimal lat,long and ODN altitude to WGS84 decimal
     *lat, long and altitude. Returns a comma separated string.
     *Sets the NewLatitude, NewLongitude and NewAltitude properties.
     *S latitudes and W longitudes are -ve.
     */
    public String OSGB36toWGS84(double OSGBlat, double OSGBlong, double OSGBalt)
    {
        //first off convert to radians
        double radWGlat = OSGBlat * deg2rad;
        double radWGlon = OSGBlong * deg2rad;

        /* these calculations were derived from the work of
         * Roger Muggleton (http://www.carabus.co.uk/) */

        /* quoting Roger Muggleton :-
         * There are many ways to convert data from one system to another, the most accurate 
         * being the most complex! For this example I shall use a 7 parameter Helmert 
         * transformation.
         * The process is in three parts: 
         * (1) Convert latitude and longitude to Cartesian coordinates (these also include height 
         * data, and have three parameters, X, Y and Z). 
         * (2) Transform to the new system by applying the 7 parameters and using a little maths.
         * (3) Convert back to latitude and longitude.
         * For the example we shall transform a GRS80 location to Airy, e.g. a GPS reading to 
         * the Airy spheroid.
         * The following code converts latitude and longitude to Cartesian coordinates. The 
         * input parameters are: WGS84 latitude and longitude, axis is the GRS80/WGS84 major 
         * axis in metres, ecc is the eccentricity, and height is the height above the 
         *  ellipsoid.
         *  v = axis / (Math.sqrt (1 - ecc * (Math.pow (Math.sin(lat), 2))));
         *  x = (v + height) * Math.cos(lat) * Math.cos(lon);
         * y = (v + height) * Math.cos(lat) * Math.sin(lon);
         * z = ((1 - ecc) * v + height) * Math.sin(lat);
         * The transformation requires the 7 parameters: xp, yp and zp correct the coordinate 
         * origin, xr, yr and zr correct the orientation of the axes, and sf deals with the 
         * changing scale factors. */

        //these are the values for OSGB36(Airy)/ODN to WGS86(GRS80)  
        double a2 = 6378137;              // WGS84_AXIS
        double e2 = 0.00669438037928458;  // WGS84_ECCENTRIC
        double h = OSGBalt;               // height above datum  (from GPS GGA sentence)
        double a = 6377563.396;         //OSGB_AXIS
        double e = 0.0066705397616;     // OSGB_ECCENTRIC 
        double xp = 446.448;
        double yp = -125.157;
        double zp = 542.06;
        double xr = 0.1502;
        double yr = 0.247;
        double zr = 0.8421;
        double s = -20.4894;

        // convert to cartesian; lat, lon are radians
        double sf = s * 0.000001;
        double v = a / (Math.sqrt(1 - (e * (Math.sin(radWGlat) * Math.sin(radWGlat)))));
        double x = (v + h) * Math.cos(radWGlat) * Math.cos(radWGlon);
        double y = (v + h) * Math.cos(radWGlat) * Math.sin(radWGlon);
        double z = ((1 - e) * v + h) * Math.sin(radWGlat);

        // transform cartesian
        double xrot = (xr / 3600) * deg2rad;
        double yrot = (yr / 3600) * deg2rad;
        double zrot = (zr / 3600) * deg2rad;
        double hx = x + (x * sf) - (y * zrot) + (z * yrot) + xp;
        double hy = (x * zrot) + y + (y * sf) - (z * xrot) + yp;
        double hz = (-1 * x * yrot) + (y * xrot) + z + (z * sf) + zp;

        // Convert back to lat, lon
        double newLon = Math.atan(hy / hx);
        double p = Math.sqrt((hx * hx) + (hy * hy));
        double newLat = Math.atan(hz / (p * (1 - e2)));
        v = a2 / (Math.sqrt(1 - e2 * (Math.sin(newLat) * Math.sin(newLat))));
        double errvalue = 1.0;
        double lat0 = 0;
        while (errvalue > 0.001)
        {
            lat0 = Math.atan((hz + e2 * v * Math.sin(newLat)) / p);
            errvalue = Math.abs(lat0 - newLat);
            newLat = lat0;
        }

        deciAlt = p / Math.cos(lat0) - v;

        //convert back to degrees
        newLat = newLat * rad2deg;
        deciLat = newLat;
        newLon = newLon * rad2deg;
        deciLon = newLon;

        return Double.toString(deciLat) + ',' + Double.toString(deciLon);
    }

	/**
	 *Converts OSGB36 decimal lat and long to an OS national grid reference.
	 *Sets the Northing and Easting properties.
	 */
    public String OSGB36toNGR(double OSGBlat, double OSGBlong)
    {
        deciLat = OSGBlat;
        deciLon = OSGBlong;
        double phi = OSGBlat * deg2rad;          // convert latitude to radians
        double lam = OSGBlong * deg2rad;          // convert longitude to radians
        double a = 6377563.396;              // OSGB semi-major axis
        double b = 6356256.91;               // OSGB semi-minor axis
        double e0 = 400000;                  // easting of false origin
        double n0 = -100000;                 // northing of false origin
        double f0 = 0.9996012717;            // OSGB scale factor on central meridian
        double e2 = 0.0066705397616;         // OSGB eccentricity squared
        double lam0 = -0.034906585039886591; // OSGB false east
        double phi0 = 0.85521133347722145;   // OSGB false north
        double af0 = a * f0;
        double bf0 = b * f0;

        // easting
        double slat2 = Math.sin(phi) * Math.sin(phi);
        double nu = af0 / (Math.sqrt(1 - (e2 * (slat2))));
        double rho = (nu * (1 - e2)) / (1 - (e2 * slat2));
        double eta2 = (nu / rho) - 1;
        double p = lam - lam0;
        double IV = nu * Math.cos(phi);
        double clat3 = Math.pow(Math.cos(phi), 3);
        double tlat2 = Math.tan(phi) * Math.tan(phi);
        double V = (nu / 6) * clat3 * ((nu / rho) - tlat2);
        double clat5 = Math.pow(Math.cos(phi), 5);
        double tlat4 = Math.pow(Math.tan(phi), 4);
        double VI = (nu / 120) * clat5 * ((5 - (18 * tlat2)) + tlat4 + (14 * eta2) - (58 * tlat2 * eta2));
        double east = e0 + (p * IV) + (Math.pow(p, 3) * V) + (Math.pow(p, 5) * VI);

        // northing
        double n = (af0 - bf0) / (af0 + bf0);
        double M = Marc(bf0, n, phi0, phi);
        double I = M + (n0);
        double II = (nu / 2) * Math.sin(phi) * Math.cos(phi);
        double III = ((nu / 24) * Math.sin(phi) * Math.pow(Math.cos(phi), 3)) * (5 - Math.pow(Math.tan(phi), 2) + (9 * eta2));
        double IIIA = ((nu / 720) * Math.sin(phi) * clat5) * (61 - (58 * tlat2) + tlat4);
        double north = I + ((p * p) * II) + (Math.pow(p, 4) * III) + (Math.pow(p, 6) * IIIA);

        // make whole number values
        east = Math.round(east);   // round to whole number
        north = Math.round(north); // round to whole number

        easting = east;
        northing = north;
        // convert to nat grid ref
         return NE2NGR(east, north);
    }

    // 
	/**
     *Converts WGS84 decimal lat, long and altitue to OSGB36 decimal
     *lat,decimal long and ODN altitude. Returns a comma separated string.
     *Sets the NewLatitude, NewLongitude and NewAltitude properties.
     *S latitudes and W longitudes are -ve.
     */
    public String WGS84toOSGB36(double WGSlat, double WGSlong, double WGSalt)
    {
        //first off convert to radians
        double radWGlat = WGSlat * deg2rad;
        double radWGlon = WGSlong * deg2rad;

        /* these calculations were derived from the work of
         * Roger Muggleton (http://www.carabus.co.uk/) */

        /* quoting Roger Muggleton :-
         * There are many ways to convert data from one system to another, the most accurate 
         * being the most complex! For this example I shall use a 7 parameter Helmert 
         * transformation.
         * The process is in three parts: 
         * (1) Convert latitude and longitude to Cartesian coordinates (these also include height 
         * data, and have three parameters, X, Y and Z). 
         * (2) Transform to the new system by applying the 7 parameters and using a little maths.
         * (3) Convert back to latitude and longitude.
         * For the example we shall transform a GRS80 location to Airy, e.g. a GPS reading to 
         * the Airy spheroid.
         * The following code converts latitude and longitude to Cartesian coordinates. The 
         * input parameters are: WGS84 latitude and longitude, axis is the GRS80/WGS84 major 
         * axis in metres, ecc is the eccentricity, and height is the height above the 
         *  ellipsoid.
         *  v = axis / (Math.sqrt (1 - ecc * (Math.pow (Math.sin(lat), 2))));
         *  x = (v + height) * Math.cos(lat) * Math.cos(lon);
         * y = (v + height) * Math.cos(lat) * Math.sin(lon);
         * z = ((1 - ecc) * v + height) * Math.sin(lat);
         * The transformation requires the 7 parameters: xp, yp and zp correct the coordinate 
         * origin, xr, yr and zr correct the orientation of the axes, and sf deals with the 
         * changing scale factors. */

        //these are the values for WGS86(GRS80) to OSGB36(Airy)/ODN
        double a = 6378137;              // WGS84_AXIS
        double e = 0.00669438037928458;  // WGS84_ECCENTRIC
        double h = WGSalt;               // height above datum  (from GPS GGA sentence)
        double a2 = 6377563.396;         //OSGB_AXIS
        double e2 = 0.0066705397616;     // OSGB_ECCENTRIC 
        double xp = -446.448;
        double yp = 125.157;
        double zp = -542.06;
        double xr = -0.1502;
        double yr = -0.247;
        double zr = -0.8421;
        double s = 20.4894;

        // convert to cartesian; lat, lon are radians
        double sf = s * 0.000001;
        double v = a / (Math.sqrt(1 - (e * (Math.sin(radWGlat) * Math.sin(radWGlat)))));
        double x = (v + h) * Math.cos(radWGlat) * Math.cos(radWGlon);
        double y = (v + h) * Math.cos(radWGlat) * Math.sin(radWGlon);
        double z = ((1 - e) * v + h) * Math.sin(radWGlat);

        // transform cartesian
        double xrot = (xr / 3600) * deg2rad;
        double yrot = (yr / 3600) * deg2rad;
        double zrot = (zr / 3600) * deg2rad;
        double hx = x + (x * sf) - (y * zrot) + (z * yrot) + xp;
        double hy = (x * zrot) + y + (y * sf) - (z * xrot) + yp;
        double hz = (-1 * x * yrot) + (y * xrot) + z + (z * sf) + zp;

        // Convert back to lat, lon
        double newLon = Math.atan(hy / hx);
        double p = Math.sqrt((hx * hx) + (hy * hy));
        double newLat = Math.atan(hz / (p * (1 - e2)));
        v = a2 / (Math.sqrt(1 - e2 * (Math.sin(newLat) * Math.sin(newLat))));
        double errvalue = 1.0;
        double lat0 = 0;
        while (errvalue > 0.001)
        {
            lat0 = Math.atan((hz + e2 * v * Math.sin(newLat)) / p);
            errvalue = Math.abs(lat0 - newLat);
            newLat = lat0;
        }

        deciAlt = p / Math.cos(lat0) - v;

        //convert back to degrees
        newLat = newLat * rad2deg;
        deciLat = newLat;
        newLon = newLon * rad2deg;
        deciLon = newLon;

        return Double.toString(deciLat) + ',' + Double.toString(deciLon);
    }


	//public properties	
    
	/**
	 *Returns the version number of the library.
	 */
	public double getVersion()
	{
		return version;
	}

 	/**
	 *Returns the Easting value resulting from the latest
	 *NGRtoOSGB36 or OSGB36toNGR transformation.
	 */
   public double getEasting()
    { return easting; }

	/**
	 *Returns the Altitude value resulting from the latest
	 *OSGB36toWGS84 or WGS84toOSGB36 transformation.
	 */
   public double getNewAltitude()
    { return deciAlt; }

	/**
	 *Returns the Latitude value resulting from the latest
	 *OSGB36toWGS84 or WGS84toOSGB36 transformation.
	 */
    public double getNewLatitude()
    {  return deciLat; }

	/**
	 *Returns the Longitude value resulting from the latest
	 *OSGB36toWGS84 or WGS84toOSGB36 transformation.
	 */
   public double getNewLongitude()
    { return deciLon; }

  	/**
	 *Returns the Northing value resulting from the latest
	 *NGRtoOSGB36 or OSGB36toNGR transformation.
	 */
   public double getNorthing()
    { return northing; }
   

    // a function used in NGR to OSGB36
    private double InitialLat(double north, double n0, double af0, double phi0, double n, double bf0)
    {
        double phi1 = ((north - n0) / af0) + phi0;
        double M = Marc(bf0, n, phi0, phi1);
        double phi2 = ((north - n0 - M) / af0) + phi1;
        double ind = 0;
        while ((Math.abs(north - n0 - M) > 0.00001) && (ind < 20))  // max 20 iterations in case of error
        {  
            ind = ind + 1;
            phi2 = ((north - n0 - M) / af0) + phi1;
            M = Marc(bf0, n, phi0, phi2);
            phi1 = phi2;
        }
        return(phi2);  
    }

    // a function used in OSGB36toNGR and NGR to OSGB36
    private double Marc(double bf0, double n, double phi0, double phi)
    {
        return bf0 * (((1 + n + ((5 / 4) * (n * n)) + ((5 / 4) * (n * n * n))) * (phi - phi0))
           - (((3 * n) + (3 * (n * n)) + ((21 / 8) * (n * n * n))) * (Math.sin(phi - phi0)) * (Math.cos(phi + phi0)))
           + ((((15 / 8) * (n * n)) + ((15 / 8) * (n * n * n))) * (Math.sin(2 * (phi - phi0))) * (Math.cos(2 * (phi + phi0))))
           - (((35 / 24) * (n * n * n)) * (Math.sin(3 * (phi - phi0))) * (Math.cos(3 * (phi + phi0)))));
    }

    //convert 12 (6e & 6n) figure numeric to letter and number grid system
    private String NE2NGR(double east, double north)
    {
        double eX = east / 500000;
        double nX = north / 500000;
        double tmp = Math.floor(eX) - 5.0 * Math.floor(nX) + 17.0;  //Math.floor Returns the largest integer less than or equal to the specified number.
        nX = 5 * (nX - Math.floor(nX));
        eX = 20 - 5.0 * Math.floor(nX) + Math.floor(5.0 * (eX - Math.floor(eX)));
        if (eX > 7.5) eX = eX + 1;
        if (tmp > 7.5) tmp = tmp + 1;
        String eing = "0000" + Double.toString(east);
        String ning = "0000" + Double.toString(north);
        // padding ensures minimum length of 5 characters
        int start = eing.length() - 7;
        int end = start + 5;
        eing = eing.substring(start, end);
        start = ning.length() - 7;
        end = start + 5;
        ning = ning.substring(start, end);
        return Character.toString((char)(tmp + 65)) + Character.toString((char)(eX + 65)) + eing + ning;

    }


}
